/*
@copyright(disclaimer)

Licensed Materials - Property of IBM
IBM DB2 Content Manager Enterprise Edition V8  (program number 5724-B19)
IBM DB2 Content Manager Express Edition V8 (program number 5724-F73)
IBM DB2 Content Manager for z/OS V8 (program number 5697-H60)
(c ) Copyright IBM Corp. 2005. All Rights Reserved.

US Government Users Restricted Rights
Use, duplication or disclosure restricted by GSA ADP Schedule
Contract with IBM Corp.

DISCLAIMER OF WARRANTIES :

Permission is granted to copy and modify this  Sample code, and to
distribute modified versions provided that both the copyright
notice,- and this permission notice and warranty disclaimer  appear in all copies and modified
versions.

THIS SAMPLE CODE IS LICENSED TO YOU AS-IS.
IBM  AND ITS SUPPLIERS AND LICENSORS  DISCLAIM
ALL WARRANTIES, EITHER EXPRESS OR IMPLIED, IN SUCH SAMPLE CODE, INCLUDING THE WARRANTY
OF NON-INFRINGEMENT AND THE IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS FOR A
PARTICULAR PURPOSE. IN NO EVENT WILL IBM OR ITS LICENSORS OR SUPPLIERS BE LIABLE FOR ANY
DAMAGES ARISING OUT OF THE USE OF  OR INABILITY TO USE THE SAMPLE CODE, DISTRIBUTION OF
THE SAMPLE CODE, OR COMBINATION OF THE SAMPLE CODE WITH ANY OTHER CODE. IN NO EVENT SHALL
IBM OR ITS LICENSORS AND SUPPLIERS BE LIABLE FOR ANY LOST REVENUE, LOST PROFITS OR DATA,
OR FOR DIRECT, INDIRECT, SPECIAL, CONSEQUENTIAL,INCIDENTAL OR PUNITIVE DAMAGES, HOWEVER
CAUSED AND REGARDLESS OF THE THEORY OF LIABILITY,-, EVEN IF IBM OR ITS LICENSORS OR SUPPLIERS
HAVE BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.

@endCopyright
*/

package com.ibm.bpe.cm.api;

import java.util.List;
import java.util.Map;
import java.util.TimeZone;

import javax.naming.InitialContext;
import javax.naming.NamingException;
import javax.sql.DataSource;

import com.ibm.bpe.api.LocalBusinessFlowManagerHome;
import com.ibm.bpe.cm.CMBPCConstants;
import com.ibm.bpe.cm.util.CollectionPoint;
import com.ibm.bpe.cm.util.CollectionPointCallbackInfo;
import com.ibm.bpe.cm.util.CollectionPointInfo;
import com.ibm.bpe.cm.util.CombinedQuery;
import com.ibm.bpe.cm.util.CombinedQueryResultSet;
import com.ibm.bpe.cm.util.ContentManagerAccess;
import com.ibm.bpe.cm.util.IntegrationException;
import com.ibm.mm.sdk.common.DKException;
import com.ibm.mm.sdk.common.DKPidICM;
import com.ibm.websphere.bo.BOFactory;
import com.ibm.websphere.sca.ServiceManager;
import commonj.sdo.DataObject;

/**
 * This class provides access to the Content Manager / Process Choreographer
 * integration function. It implements the "business delegate" pattern to
 * hide any complexities involved with lookup of an EJB which eventually contains
 * the real implementation.<br> 
 * Use these static functions inside a Process Choreographer Java snippet
 * or inside a Java transition condition.<br>
 * The integration code makes some assumptions about process attributes,
 * the most important one is the Content Manager itemPID of an associated
 * document or document folder. Use <code>initializeCustomProperties</code>
 * at the very beginning of your business process to set these properties properly.<br>
 */
public class CMBPCIntegration {

	public static String getPid(String itemId,
								String componentId,
								Integer versionNo) {
		String result = null;
		ContentManagerAccess cma = getCMA();
		if (cma != null) {
			result = cma.getPid(itemId, componentId, versionNo.intValue());
		}
		return result;
	}

	/**
	* Evaluate a Content Manager XPath expression.
	* Use this function in a Process Choreographer transition condition to
	* determine a flow route based on document contents.
	* @param xPathQueryString A Content Manager XPath expression for a condition
	* @return true, if the condition is true
	*/
	public static boolean evaluateCondition(String xPathQueryString) {
		boolean result = false;
		ContentManagerAccess cma = getCMA();
		if (cma != null) {
	       	result = cma.evaluateCondition(xPathQueryString);
	    }
		return result;
	}

	/**
	* Get an attribute value of the associated item (when it is no folder).
	* @param pidString - PID string representing a Content Manager item
	* @param attributeSpec - Attribute to retrieve in &lt;item-type&gt;/&lt;attribute-name&gt; 
	* format, e.g. <code>"XYZ_ClaimForm/XYZ_DriversLic"</code>.
	* Note that &lt;attribute-name&gt; can include child component names,
	* e.g. Book/Chapter/Section/Author.
	* @return an Object for the attribute value or null if the attribute cannot be resolved.
	*/
	public static Object getContentAttribute(String pidString,
									  String attributeSpec) {
		Object value = null;
		ContentManagerAccess cma = getCMA();
		if (cma != null) {
	       	value = cma.getContentAttribute(pidString, attributeSpec);
	    }
		return value;
	}

   /**
    * Get the attributes of of the specified item.
  	* @param pidString - PID string representing a Content Manager item
	* @param attributeSpecArray - array of attributes to retrieve in &lt;item-type&gt;/&lt;attribute-name&gt;
	* format, e.g.
	* <code><pre>
	* {
	*     "XYZ_ClaimForm/XYZ_DriversLic",
	*     "XYZ_ClaimForm/XYZ_PolicyNum",
	*     "XYZ_AdjReport/XYZ_LicPlate"
	* }
	* </pre></code>
	* Note that &lt;attribute-name&gt; can include child component names,
	* e.g. Book/Chapter/Section/Author.
   * @return a map that maps attribute specs to the corresponding values
	*/
	public static Map getContentAttributes(String pidString,
									      String[] attributeSpecArray) {
		Map valueMap = null;
		ContentManagerAccess cma = getCMA();
		if (cma != null) {
	       	valueMap = cma.getContentAttributes(pidString, attributeSpecArray);
	    }
		return valueMap;
	}

   /**
  	* Get an attribute value of items in the specified folder.
  	* This is useful when a CM folder is started on a WPC
  	* process, and the process needs to get the value of
  	* attributes of items (e.g. the license plate attribute
  	* in the adjuster's report item type).
  	* @param pidString - PID string representing a Content Manager item
  	* @param attributeSpec - Attribute to retrieve in &lt;item-type&gt;/&lt;attribute-name&gt; 
  	* format, e.g. <code>"XYZ_ClaimForm/XYZ_DriversLic"</code>.
  	* Note that &lt;attribute-name&gt; can include child component names,
  	* e.g. Book/Chapter/Section/Author.
  	* @return an Object for the attribute value or null if the attribute cannot be resolved.		
  	*/
	public static Object getFolderContentAttribute(String pidString,
				  							String attributeSpec) {
		Object value = null;
		ContentManagerAccess cma = getCMA();
		if (cma != null) {
	       	value = cma.getFolderContentAttribute(pidString, attributeSpec);
	    }
		return value;
	}

   /**
    * Get the attributes of items in the specified folder.
    * This is useful when a CM folder is started on a WPC
    * process, and the process needs to get the value of
    * attributes of items (e.g. the license plate attribute
    * in the adjuster's report item type).
  	* @param pidString - PID string representing a Content Manager item
	* @param attributeSpecArray - array of attributes to retrieve in &lt;item-type&gt;/&lt;attribute-name&gt;
    * format, e.g.
    * <code><pre>
    * {
    *     "XYZ_ClaimForm/XYZ_DriversLic",
    *     "XYZ_ClaimForm/XYZ_PolicyNum",
    *     "XYZ_AdjReport/XYZ_LicPlate"
    * }
    * </pre></code>
    * Note that &lt;attribute-name&gt; can include child component names,
    * e.g. Book/Chapter/Section/Author.
   * @return a map that maps attribute specs to the corresponding values
    */
	public static Map getFolderContentAttributes(String pidString,
					   						     String[] attributeSpecArray) {
		Map valueMap = null;
		ContentManagerAccess cma = getCMA();
		if (cma != null) {
	       	valueMap = cma.getFolderContentAttributes(pidString, attributeSpecArray);
	    }
		return valueMap;
	}

	/**
	 * Run a combined query for Process Choreographer and Content Manager data.
	 * For more details on the QueryResultSet and how to specify the clauses, also refer
	 * to the Process Choreographer documentation.<br>
	 * <b>Note:</b> If you plan to use this function from an EJB or Servlet context, consider 
	 * calling <code>combinedQuery</code> on the <code>ContentManagerIntegrationEjb</code>
	 * instead using local EJB calls. This may give you better performance than through this
	 * business delegate class.<br>
	 *   
	 * @param selectClauseWorkflow The requested Process Choreographer column names. Must not be null.
	 * @param flowItemTypeNameContent The item type of the item which is routed through the flow
	 * This can be a document item type or a folder item type. Must not be null.
	 * @param flowItemTypeAttributesContent A comma separated list of attributes of the flow item type
	 * that should be returned in the result set. 
	 * @param additionalAttributesContent If the flowItemTypeContent is a folder, this specifies an 
	 * optional list of attributes of the contained documents. The list is comma separated and
	 * the attributes are specified in the form <code>&lt;itemtype&gt;/&lt;attribute&gt;</code>.   
	 * @param whereClauseWorkflow The SQL where condition for Process Choreographer data, may be null.
	 * @param xPathContent  An valid XPath expression to filter the Content Manager data. The result must
	 * be the flow item type. To express conditions on contained item attributes make sure
	 * to use an inboundlink to get to the folder item. 	  	
	 * @param orderByClauseWorkflow The order clause for Process Choreographer data, may be null.
	 * @param timezoneWorkflow A timezone specification for timestamp given or returned, may be null. 
	 * @return The result of the combined query.
	 */
	
	public static CombinedQueryResultSet combinedQuery(String selectClauseWorkflow,
			String flowItemTypeNameContent,
			String flowItemTypeAttributesContent,
			String additionalAttributesContent,
			String whereClauseWorkflow,
			String xPathContent,
			String orderByClauseWorkflow,
			String timeZoneIDWorkflow) {
		
		CombinedQueryResultSet qrs = null;
		TimeZone timeZoneWorkflow = null;
		if( timeZoneIDWorkflow != null )
		{
			timeZoneWorkflow = TimeZone.getTimeZone(timeZoneIDWorkflow);			
		}
					
		CombinedQuery cq = getCQ();
		if (cq != null) {
			qrs = cq.combinedQuery(selectClauseWorkflow, flowItemTypeNameContent, flowItemTypeAttributesContent,
								   additionalAttributesContent, whereClauseWorkflow,
								   xPathContent, orderByClauseWorkflow, timeZoneWorkflow);
		}
		return qrs;
	}
	

	   /**
		* Add the results of a Content Manager query to the folder that is routed through the process.
		* 
		* @param pidString - PID string representing a Content Manager item
		* @param query XPath query string that specifies the items to be added
		* @return number of items that have been added to the folder (number of hits); -1 in case of failure
		*/
		public static int addItemToFlowFolder(String pidString,
										   		  String xPathQueryString) {
			int result = -1;
			ContentManagerAccess cma = getCMA();
			if (cma != null) {
		       	result = cma.addItemToFlowFolder(pidString, xPathQueryString);
		    }
			return result;
		}

	   /**
	    * Add the item routed through the process to the folder retrieved by a Content Manager query. 
	    * 
	    * @param pidString - PID string representing a Content Manager item
	    * @param query XPath query string that specifies the folder to which this item should be added
	    * @return number of folders that matched the query; -1 in case of failure
	    */
		public static int addFlowItemToFolder(String pidString,
										   String xPathQueryString) {
			int result = -1;
			ContentManagerAccess cma = getCMA();
			if (cma != null) {
		       	result = cma.addFlowItemToFolder(pidString, xPathQueryString);
		    }
			return result;
		}

	   /**
	    * Assign the item routed through the process to a new item type.
	    * Note: this changes the itemPid so it needs to be reassigned to the process
	    * property when reindex() has been called to ensure the client displays
	    * the appropriate set of attributes/parts.
	    *   
	    * @param pidString - PID string representing a Content Manager item
	    * @param newItemType name of the new item type the item should be assigned to
	    * @return number of folders that matched the query
	    */
		public static void reindex(String pidString,
						    String newItemType) {
			ContentManagerAccess cma = getCMA();
			if (cma != null) {
		       	cma.reindex(pidString, newItemType);
		    }
		}

	   /**
		* Set the ACL of an item to a new value. Note that this requires item-level ACL checking.
		* 
	    * @param pidString - PID string representing a Content Manager item
		* @param newACL - ACL to be assigned to this item
		*/
		  
		public static void setACL(String pidString,
						   String newACL) {
			ContentManagerAccess cma = getCMA();
			if (cma != null) {
		       	cma.setACL(pidString, newACL);
		    }
		}
		
		/**
		 * Start a collection point condition.
		 * @param callbackInfo Information to call back to the requester
		 * @param collectionPointRequest Data object containing an item reference, collection point
		 * condition and timeout values
		 */
		public static void collectionPoint( CollectionPointCallbackInfo callbackInfo, DataObject collectionPointRequest )
		{
			CollectionPoint cp = getCollectionPoint();
			DataObject itemRef = collectionPointRequest.getDataObject(CMBPCConstants.BOATT_COLLECTION_POINT_REQUEST_ITEMREF);
			String pidString = itemRef.getString(CMBPCConstants.BOATT_ITEM_REF_PIDSTRING);
			try {
				DKPidICM pid = new DKPidICM(pidString);
				String itemId = pid.getItemId();

				// Parse the condition data object and move information 
				// to itemtype names and item quantity arrays
				DataObject condition = collectionPointRequest.getDataObject(CMBPCConstants.BOATT_COLLECTION_POINT_REQUEST_CONDITION);
				List requiredItems = condition.getList(CMBPCConstants.BOATT_COLLECTION_POINT_REQUEST_CONDITION_RI);
				int numReqItems = requiredItems.size();
				String [] itemTypeNames = new String[numReqItems];
				int [] itemTypeQuantities = new int [numReqItems];
				for(int i = 0; i < numReqItems; i++) {
					DataObject itemCondition = (DataObject) requiredItems.get(i);
					itemTypeNames[i] = itemCondition.getString(CMBPCConstants.BOATT_COLLECTION_POINT_REQUEST_CONDITION_RI_ITEMTYPE);
					itemTypeQuantities[i] = itemCondition.getInt(CMBPCConstants.BOATT_COLLECTION_POINT_REQUEST_CONDITION_RI_ITEMQUANTITY);				
				}
				int timeout = collectionPointRequest.getInt(CMBPCConstants.BOATT_COLLECTION_POINT_REQUEST_TIMEOUT);
				CollectionPointInfo cpInfo = new CollectionPointInfo(
						callbackInfo, itemId, itemTypeNames, itemTypeQuantities, (long) timeout);
							
				cp.addCollectionPointRequest( cpInfo );							
			}
			catch (DKException e) {
				// a DKException might occur in case of an invalid itemPid
				throw new IntegrationException(e);
			}
		}

	/**
	* Internal function to cache a private instance of the ContentManagerAccess class.
	* @return instance of ContentManagerAccess
	*/
	static private ContentManagerAccess _cma = null;
	
	static private ContentManagerAccess getCMA() {
			if (_cma == null) _cma = new ContentManagerAccess();
			return _cma;
	}
	
	/**
	* Internal function to cache a private instance of the CombinedQuery class.
	* @return instance of CombinedQuery
	*/
	static private CombinedQuery _cq = null;
	
	static private CombinedQuery getCQ() {
		if (_cq == null) {								 		
				_cq = new CombinedQuery(getLocalBusinessFlowManagerHome(), getBOF());
		}		
		return _cq;
	}
	
	/**
	* Internal function to cache a private instance of the CollectionPoint class.
	* @return instance of CollectionPoint
	*/
	static private CollectionPoint _cp = null;
	
	static private CollectionPoint getCollectionPoint() {
		if (_cp == null) {								 	
   		  _cp = new CollectionPoint(getLsDataSource());
		}
		return _cp;
	}
	
	/**
	* Internal function to cache the Business Object Factory
	* @return business object factory
	*/
	static private BOFactory _bof = null;

	static private BOFactory getBOF() {
		if (_bof == null)_bof = (BOFactory) ServiceManager.INSTANCE.locateService(CMBPCConstants.BO_FACTORY);
		return _bof;
	}
	
	
	/**
	* Internal function to cache a refernce to the local home i/f for the
	* business flow manager.
	*/
	static private LocalBusinessFlowManagerHome _bfmHome = null;
	
	/**
	 * Lookup the local home interface of the BusinessFlowManager bean
	 * Internal function to cache a refernce to the local home i/f for the
     * business flow manager.
	 * @return instance of the local home interface
	 */
	static private LocalBusinessFlowManagerHome getLocalBusinessFlowManagerHome() {

		if (_bfmHome == null) {
			try {
				InitialContext ic = new InitialContext();
				_bfmHome =
					 (LocalBusinessFlowManagerHome) ic.lookup(CMBPCConstants.LOCAL_BMF_HOME_REF);				
			}
			catch (NamingException ne) {
				throw new IntegrationException(ne);
			}
		}
		return _bfmHome;
	}	
	
	/**
	* Internal function to cache the data source
	* for the CM library server 
	*/
	static private DataSource _lsDataSource = null;
	
	/**
	* Internal function to cache the data source
	* for the CM library server 
	* @return DataSource for CM LS
	*/
	static private DataSource getLsDataSource()
	{
		if( _lsDataSource == null )
		{
			try {
				InitialContext ic = new InitialContext();
				_lsDataSource = (DataSource) ic.lookup("jdbc/ICMNLSDB");
			}
			catch (NamingException ne) {
				throw new IntegrationException(ne);
			}						
		}
		return _lsDataSource;
	}
	

}
