/*
@copyright(disclaimer)

Licensed Materials - Property of IBM
IBM DB2 Content Manager Enterprise Edition V8  (program number 5724-B19)
IBM DB2 Content Manager Express Edition V8 (program number 5724-F73)
IBM DB2 Content Manager for z/OS V8 (program number 5697-H60)
(c ) Copyright IBM Corp. 2005. All Rights Reserved.

US Government Users Restricted Rights
Use, duplication or disclosure restricted by GSA ADP Schedule
Contract with IBM Corp.

DISCLAIMER OF WARRANTIES :

Permission is granted to copy and modify this  Sample code, and to
distribute modified versions provided that both the copyright
notice,- and this permission notice and warranty disclaimer  appear in all copies and modified
versions.

THIS SAMPLE CODE IS LICENSED TO YOU AS-IS.
IBM  AND ITS SUPPLIERS AND LICENSORS  DISCLAIM
ALL WARRANTIES, EITHER EXPRESS OR IMPLIED, IN SUCH SAMPLE CODE, INCLUDING THE WARRANTY
OF NON-INFRINGEMENT AND THE IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS FOR A
PARTICULAR PURPOSE. IN NO EVENT WILL IBM OR ITS LICENSORS OR SUPPLIERS BE LIABLE FOR ANY
DAMAGES ARISING OUT OF THE USE OF  OR INABILITY TO USE THE SAMPLE CODE, DISTRIBUTION OF
THE SAMPLE CODE, OR COMBINATION OF THE SAMPLE CODE WITH ANY OTHER CODE. IN NO EVENT SHALL
IBM OR ITS LICENSORS AND SUPPLIERS BE LIABLE FOR ANY LOST REVENUE, LOST PROFITS OR DATA,
OR FOR DIRECT, INDIRECT, SPECIAL, CONSEQUENTIAL,INCIDENTAL OR PUNITIVE DAMAGES, HOWEVER
CAUSED AND REGARDLESS OF THE THEORY OF LIABILITY,-, EVEN IF IBM OR ITS LICENSORS OR SUPPLIERS
HAVE BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.

@endCopyright
*/

package com.ibm.bpe.cm.util;
import javax.naming.InitialContext;
import javax.naming.NamingException;

import com.ibm.bpe.cm.CMBPCConstants;
import com.ibm.bpe.cm.conmgmt.Environment;
import com.ibm.websphere.asynchbeans.Work;
import com.ibm.websphere.asynchbeans.WorkException;
import com.ibm.websphere.asynchbeans.WorkItem;
import com.ibm.websphere.asynchbeans.WorkManager;

/**
 * The working class of an Async Bean which loops in the run()
 * method and performs the necessary checks on the CM Integration EJB.
 */
public class AsyncBeanWork implements Work
{

  /**
   * Constructor which takes a ContentEventHandling and a CollectionPoint.
   * Both are used to perform the appropriate operations if a Content Event occurs or
   * a Collection Point is started.
   * @param ceh ContentEventHandling
   * @param cp CollectionPoint
   */
  public AsyncBeanWork(ContentEventHandling ceh, CollectionPoint cp)
  {
  	_ceh = ceh;
    _cp = cp;
  }

  public void release()
  {
  }

  /**
   * This is the main loop of the Async Beans where we run
   * multiple checks (such as if the collection point is ready).
   * The checking code itself is run in the context of the
   * EJB (we require a valid transactional context).	 
   * @see java.lang.Runnable#run()
   */
  public void run()
  {
    do
    {
      try
      {
      	//
      	// Check for process instances to be started/terminated
      	//
      	_ceh.checkContentEvent();
        //
        // Run the check for collection points
        //
        _cp.checkCollectionPoint();
      }
      catch (Throwable exc)
      {
        //
        // It is important to catch all kind of exceptions
        // here in order to avoid that this thread terminates.
        //
        exc.printStackTrace();
      }
      // After all requests have been processed, sleep some time
      // before querying again. 500 ms shouldn't harm overall performance
      // if the check is fast enough.
      try
      {
        Thread.sleep(_sleepTime);
      }
      catch (InterruptedException exc)
      {
        exc.printStackTrace();
      }
    }
    while (_running);
  }

  /**
   * Stop looping in run().
   * We probably don't need to call this since the reference
   * to this AsyncBean is stored in a static variable of the parent EJB.
   */
  static public void stopRunning()
  {
    _running = false;
  }

  /**
   * Flag if loop can continue to run.
   * Function "stopRunning" aborts the loop in run()	 
   */
  static boolean _running = true;

  /**
   * This function is called from the startup bean to launch the asynchronous
   * thread which is used for polling the collection point conditions etc.
   * Since this is a long-running async thread, we use the daemon option
   * for the startWork() function.
   * 
   * @param initialContext naming context for work manager lookup
   * @param ceh content event handling (to be initialized)
   * @param cp collection point (to be initialized)
   */
  static public void initAsyncBeanWork(InitialContext initialContext,
  									   ContentEventHandling ceh,
									   CollectionPoint cp)
  {
    if (_asyncBeanWorkItem == null)
    {
      try
      {
      	//
      	// Check for a non-default sleep time
      	//
      	String sCheckFrequency = Environment.getProperty(CHECK_FREQUENCE_PROPERTY );
      	if( sCheckFrequency != null )
      	{
      		int checkFrequency = Integer.parseInt( sCheckFrequency );
      		// Use only if "reasonable" values - between 10 ms and 10 hours
      		if( checkFrequency > 100 && checkFrequency < 36000000)
      		{
      			_sleepTime = checkFrequency;      			
      		}
      	}
      	//
      	// Use the default work manager to launch the async polling
      	// thread as a daemon.
      	//
        WorkManager wm = (WorkManager) initialContext.lookup(CMBPCConstants.WORK_MANAGER_REF);
        AsyncBeanWork asyncBeanWork = new AsyncBeanWork(ceh, cp);
        _asyncBeanWorkItem = wm.startWork(asyncBeanWork, true);
      }
      catch (NamingException exc)
      {
        throw new IntegrationException(exc);
      }
      catch (WorkException exc)
      {
      	throw new IntegrationException(exc);
      }
    }
  }
  
  /**
   * This function is called when the server is stopped or the application is undeployed.
   * It terminates the loop in the AsynchBeanWork's run() method.
   */

  static public void stopCollectionPoint()
  {
    if (_asyncBeanWorkItem != null)
    {
      _running = false;
    }
  }
  
  static WorkItem _asyncBeanWorkItem = null;

  ContentEventHandling _ceh = null;
  CollectionPoint _cp = null;

  // make the reference available to the startup bean 
  public static final String BFM_HOME_REF		= CMBPCConstants.BFM_HOME_REF;
  public static final String BPEDB_DS_REF		= CMBPCConstants.BPEDB_DS_REF;
  public static final String ICMNLSDB_DS_REF	= CMBPCConstants.ICMNLSDB_DS_REF;

  // Time (in ms) to wait for the next check of collection point information
  private static final int DEFAULT_SLEEP_TIME = 10000;
  private static int _sleepTime = DEFAULT_SLEEP_TIME;
  private static final String CHECK_FREQUENCE_PROPERTY = "ContentManagerCheckFrequency";
  
}
