/*
@copyright(disclaimer)

Licensed Materials - Property of IBM
IBM DB2 Content Manager Enterprise Edition V8  (program number 5724-B19)
IBM DB2 Content Manager Express Edition V8 (program number 5724-F73)
IBM DB2 Content Manager for z/OS V8 (program number 5697-H60)
(c ) Copyright IBM Corp. 2005. All Rights Reserved.

US Government Users Restricted Rights
Use, duplication or disclosure restricted by GSA ADP Schedule
Contract with IBM Corp.

DISCLAIMER OF WARRANTIES :

Permission is granted to copy and modify this  Sample code, and to
distribute modified versions provided that both the copyright
notice,- and this permission notice and warranty disclaimer  appear in all copies and modified
versions.

THIS SAMPLE CODE IS LICENSED TO YOU AS-IS.
IBM  AND ITS SUPPLIERS AND LICENSORS  DISCLAIM
ALL WARRANTIES, EITHER EXPRESS OR IMPLIED, IN SUCH SAMPLE CODE, INCLUDING THE WARRANTY
OF NON-INFRINGEMENT AND THE IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS FOR A
PARTICULAR PURPOSE. IN NO EVENT WILL IBM OR ITS LICENSORS OR SUPPLIERS BE LIABLE FOR ANY
DAMAGES ARISING OUT OF THE USE OF  OR INABILITY TO USE THE SAMPLE CODE, DISTRIBUTION OF
THE SAMPLE CODE, OR COMBINATION OF THE SAMPLE CODE WITH ANY OTHER CODE. IN NO EVENT SHALL
IBM OR ITS LICENSORS AND SUPPLIERS BE LIABLE FOR ANY LOST REVENUE, LOST PROFITS OR DATA,
OR FOR DIRECT, INDIRECT, SPECIAL, CONSEQUENTIAL,INCIDENTAL OR PUNITIVE DAMAGES, HOWEVER
CAUSED AND REGARDLESS OF THE THEORY OF LIABILITY,-, EVEN IF IBM OR ITS LICENSORS OR SUPPLIERS
HAVE BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.

@endCopyright
*/

package com.ibm.bpe.cm.util;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;

import javax.sql.DataSource;

import com.ibm.bpe.cm.CMBPCConstants;

/**
 * This class implements the collection point logic that lets a process wait
 * until certain numbers of items of a specified item types are stored in the
 * folder that is routed through the process.
 * 
 * See section 4.4 of the documentation.
 */

public class CollectionPoint
{
	
  /**
   * Constructor which takes a library server data source.
   * @param libraryServerDataSource
   */	
  public CollectionPoint( DataSource libraryServerDataSource )
  {    
    _lsDataSource = libraryServerDataSource;    
  }
  

  /**
   * checks if the required tables for the collection point instances
   * exists, creates them if needed
   * @throws SQLException
   */
  public void createDatabaseTable() throws SQLException
  {
    java.sql.Connection con = _lsDataSource.getConnection();
    java.sql.Statement stmt = con.createStatement();
    try
    {
      stmt.execute(CHECK_COLLECTIONPOINTS_TABLE);
    }
    catch (SQLException exc)
    {
    	String sqlState = exc.getSQLState();
    	if (sqlState.equals("42703") || sqlState.equals("42S22")) {    // incompatible columns or data types
    		System.out.println(">> Dropping old version of BPECOLLECTIONPOINTS table");
    		stmt.execute(DROP_COLLECTIONPOINTS_TABLE);      	
    	}
    	System.out.println(">> Creating BPECOLLECTIONPOINTS table");
    	stmt.execute(CREATE_COLLECTIONPOINTS_TABLE);
    }
    try
    {
    	stmt.execute(CHECK_COLLECTIONPOINTITEMTYPES_TABLE);
    }
    catch (SQLException exc)
    {
    	String sqlState = exc.getSQLState();
        if (sqlState.equals("42703") || sqlState.equals("42S22")) {    // incompatible columns or data types
        	System.out.println(">> Dropping old version of BPECOLLECTIONPOINTITEMTYPES table");
        	stmt.execute(DROP_COLLECTIONPOINTITEMTYPES_TABLE);      	
        }
        System.out.println(">> Creating BPECOLLECTIONPOINTITEMTYPES table");
        stmt.execute(CREATE_COLLECTIONPOINTITEMTYPES_TABLE);
        System.out.println(">> Creating index on BPECOLLECTIONPOINTITEMTYPES");
        stmt.execute(CREATE_COLLECTIONPOINTITEMTYPES_INDEX );
    }
    try
    {
      // Check syscat, if bpecm is already created
      ResultSet resultSet = stmt.executeQuery(CHECK_SEQUENCE);
      if( !resultSet.next())
      {
      	// No row returned means that sequence does not exist: create it
      	System.out.println(">> Creating bpecm sequence");
      	stmt.execute(CREATE_SEQUENCE);      	
      }
      resultSet.close();
    }
    catch (SQLException exc)
    {
      System.out.println(">> Creating bpecm sequence failed");      
    }    
    finally
    {
      if (con != null)
        con.close();
    }
  }

  /**
   * This function is regularly called from the AsyncBean to check if the collection points
   * stored in the database are ready.
   * The check is performed and if the condition is met, the entry is removed from the database
   * table and a notification is sent back to the business process.
   * This all runs in a global transaction.	 
   */
  
  static int count = 0;
  public void checkCollectionPoint()
  {
    // The connection - to be closed in finally
    Connection con = null;
    PreparedStatement pstmt = null;
    ResultSet rs = null;
    
    // query the database to get a list of collection points to check
    try
    {
      con = _lsDataSource.getConnection();
 
      // Select old request collectionpoints (deleted folders)
      pstmt = con.prepareStatement(QUERY_COLLECTIONPOINTS_OLDREQUESTS_TABLE);
      rs = pstmt.executeQuery();
      while (rs.next())
      {
      	int seqId = rs.getInt(1);
        byte [] correlation = rs.getBytes(2);                
        CollectionPointCallbackInfo callbackInfo = (CollectionPointCallbackInfo) deserialize( correlation );
        callbackInfo.notifyBusinessProcess( CMBPCConstants.COLLECTIONPOINT_ITEM_DELETED );        
        deleteCollectionPoint(con, seqId);
        rs.close();
        rs = pstmt.executeQuery();
      }
      pstmt.close();

       // Select the collection points with empty parameter arrays (=success)
      pstmt = con.prepareStatement(QUERY_COLLECTIONPOINTS_EMPTY_TABLE);
      rs = pstmt.executeQuery();
      
     while (rs.next())
      {
        int seqId = rs.getInt(1);
        byte [] correlation = rs.getBytes(2);
        String itemId = rs.getString(3);
        CollectionPointCallbackInfo callbackInfo = (CollectionPointCallbackInfo) deserialize( correlation );
        callbackInfo.notifyBusinessProcess( CMBPCConstants.COLLECTIONPOINT_OK );   
        deleteCollectionPoint(con, seqId);
        rs.close();
        rs = pstmt.executeQuery();
      }
      rs.close();
      pstmt.close();
 
      // Select the collection point requests to process
      PreparedStatement outerPstmt = con.prepareStatement(QUERY_COLLECTIONPOINTS_TABLE);
      ResultSet outerResultSet = outerPstmt.executeQuery();
      pstmt = con.prepareStatement( QUERY_CORRELATION );
      while (outerResultSet.next())
      {
        int seqId = outerResultSet.getInt(1);
        pstmt.setInt( 1, seqId );        
        rs = pstmt.executeQuery();
        if( rs.next() )
        {
            byte [] correlation = rs.getBytes(1);
            CollectionPointCallbackInfo callbackInfo = (CollectionPointCallbackInfo) deserialize( correlation );
            callbackInfo.notifyBusinessProcess( CMBPCConstants.COLLECTIONPOINT_OK );
            deleteCollectionPoint(con, seqId);
        }
        rs.close();
        outerResultSet = outerPstmt.executeQuery();
      }
      pstmt.close();
      outerResultSet.close();
      outerPstmt.close();

      // Select the collection points which timed out
      pstmt = con.prepareStatement(QUERY_COLLECTIONPOINTS_TIMEDOUT_TABLE);
      // get the current timestamp to check if a request is due
      pstmt.setLong(1, System.currentTimeMillis());
      rs = pstmt.executeQuery();
      while (rs.next())
      {
      	int seqId = rs.getInt(1);
        byte [] correlation = rs.getBytes(2);
        CollectionPointCallbackInfo callbackInfo = (CollectionPointCallbackInfo) deserialize( correlation );
        callbackInfo.notifyBusinessProcess( CMBPCConstants.COLLECTIONPOINT_TIMEOUT );        
        deleteCollectionPoint(con, seqId);
        rs.close();
        rs = pstmt.executeQuery();
      }
      rs.close();

    }
    catch (SQLException exc)
    {
      throw new IntegrationException(exc);
    }
    finally
    {
      if (con != null)
      {
        try
        {
            rs.close();
            pstmt.close();
            con.close();
        }
        catch (SQLException e)
        {
        }
      }
    }

  }

  /**
   * deletes all entries matching the given seqId in the collection point tables
   * @param con connection used to perform the delete query
   * @param seqId sequence Id
   * @throws SQLException
   */
  private void deleteCollectionPoint(Connection con, int seqId) throws SQLException
  {
    PreparedStatement pstmt = null;
    try
    {
      // first delete the collection point instance entry
      pstmt = con.prepareStatement(DELETE_COLLECTIONPOINTS_TABLE);
      pstmt.setInt(1, seqId );
      pstmt.executeUpdate();
      pstmt.close();
      // now delete the requirements
      pstmt = con.prepareStatement(DELETE_COLLECTIONPOINTITEMTYPES_TABLE);
      pstmt.setInt(1, seqId);
      pstmt.executeUpdate();
      pstmt.close();
    }
    finally
    {
      if (pstmt != null)
      {
        pstmt.close();
      }
    }
  }


  /**
   * Register a collection point in the database.
   * All information of the "info" object is stored in database tables, except itemTypes with
   * non-positive quantities, which are filtered out and disregarded.
   * @param info contains the information describing the collection point to be created
   */
  public void addCollectionPointRequest(CollectionPointInfo info)
  {

    Connection con = null;
    try
    {
      con = _lsDataSource.getConnection();

      String[] itemTypeNames = info.getItemTypeNames();
      int[] itemTypeQuantities = info.getItemTypeQuantities();

      // Check for null values
      if (itemTypeNames == null || itemTypeQuantities == null)
      {
        itemTypeNames = new String[] {};
        itemTypeQuantities = new int[] {};
      }

      // Check for duplicate itemTypes
      for (int i = 0; i < itemTypeNames.length; i++)
      {
        for (int j = i + 1; j < itemTypeNames.length; j++)
        {
          if (itemTypeNames[i].equals(itemTypeNames[j]))
            System.out.println(
              "!> Duplicate item types in collection point parameter, "
                + "using maximum individual quantity");
        }
      }
      
      // Retrieve new sequence ID for the collectionpoint entry
      Statement stmt = con.createStatement();
      ResultSet seqResult = stmt.executeQuery( NEXT_SEQUENCE_VALUE );
      int seqId = 0;
      if( seqResult.next() )
      {
      	seqId = seqResult.getInt( 1 );      	
      }
      else
      {
      	throw new IntegrationException("Unable to read next sequence number");
      }
      seqResult.close();
      stmt.close();
                   
      PreparedStatement pstmt = con.prepareStatement(INSERT_COLLECTIONPOINTITEMTYPES_TABLE);
      
      // Handle cases where the names and quantities arrays have differing length
      // by using the smaller length as bound
      int noOfItems = Math.min(itemTypeNames.length, itemTypeQuantities.length);

      if (itemTypeNames.length != itemTypeQuantities.length)
      {
        System.out.println("!> Collection point parameter arrays have differing length, truncating");
      }

      // Put the requirements into the tables first
      for (int i = 0; i < noOfItems; i++)
        if (itemTypeNames[i] != null && itemTypeQuantities[i] > 0) // filter nonsense quantities
        {
          pstmt.setInt(1, seqId );
          pstmt.setString(2, itemTypeNames[i]);
          pstmt.setInt(3, itemTypeQuantities[i]);
          pstmt.executeUpdate();
        }
      pstmt.close();

      // Now put the actual instance into the collection point table
      pstmt = con.prepareStatement(INSERT_COLLECTIONPOINTS_TABLE);
      pstmt.setInt(1, seqId );      
      pstmt.setBytes( 2, serialize( info.getCallbackInfo() ) );
      pstmt.setString(3, info.getFolderItemId());
      pstmt.setLong(4, System.currentTimeMillis() + (info.getTimeout() * 1000));  // condition will time out at this point in time
      pstmt.executeUpdate();
      pstmt.close();

    }
    catch (Exception exc)
    {
      throw new IntegrationException(exc);
    }
    finally
    {
      if (con != null)
      {
        try
        {
          con.close();
        }
        catch (SQLException e)
        {
        }

      }
    }
  }

  /**
   * Serialize an object to a byte array
   * @param ser A serializable object
   * @return A byte array representation of the oject
   */
  private static byte[] serialize( Serializable ser )
  {
  	try
    {
      ByteArrayOutputStream baos = new ByteArrayOutputStream();
      ObjectOutputStream oos = new ObjectOutputStream(baos);
      oos.writeObject(ser);
      oos.close() ;        
      return baos.toByteArray();
    }
    catch (IOException exc)
    {
      throw new IntegrationException(exc);
    }  	
  }
  
 /**
   * De-serialize an object from a byte array.
  *
  * @param serializedObject The byte array containing the serialized object.
  * @return
  *    The object recreated from the serialized format. Will be <code>null</code> if
  *    the serialized object parameter was <code>null</code>.
  */
  private static final Serializable deserialize(byte[] serializedObject)                      
  {
     if (serializedObject == null) return null;
     try
     {    
       ByteArrayInputStream bis = new ByteArrayInputStream(serializedObject);
       ObjectInputStream ois = new ObjectInputStream(bis);
       Serializable serializable = (Serializable) ois.readObject();
       ois.close();       
       return serializable;
     }
     catch(IOException exc)
     {
       throw new IntegrationException(exc) ;
     }
     catch( ClassNotFoundException exc )
	 {
        throw new IntegrationException(exc) ;
	 }
  }


  // Data source for the CM library server - used to query if a collection point is ready
  private DataSource _lsDataSource = null;
  

  // Check if the CMCOLLPOINT table exists
  private final String CHECK_COLLECTIONPOINTS_TABLE = 
  	"select seqid from bpecollectionpoints fetch first row only";

  // Check if the CMCOLLPOINTITEMTYPES table exists
  private final String CHECK_COLLECTIONPOINTITEMTYPES_TABLE =
    "select seqid from bpecollectionpointitemtypes fetch first row only";

  // Create the BPECM sequence (used to generate primary keys for the
  // bpecollectionpoints table
  private final String CREATE_SEQUENCE =
  	"create sequence BPECM as integer cycle";
  
  // Check if the sequence already exists
  private final String CHECK_SEQUENCE =
  	"select 1 from syscat.sequences where seqname='BPECM'";
  
  // Get next value for the sequence
  private final String NEXT_SEQUENCE_VALUE =
  	"values nextval for BPECM";
  
  // Create the bpecollectionpoints table
  private final String CREATE_COLLECTIONPOINTS_TABLE =
    "create table bpecollectionpoints "
      + "( "
      + "	seqid integer not null primary key, "
      + "	createts timestamp not null default current timestamp, "
      + "	correlation blob(1M) not null, "
      + "	itemid varchar(26) not null, "
      + "	timeout bigint not null "
      + ")";
  
  // Drop the bpecollectionpoints table if an old version has been detected
  private final String DROP_COLLECTIONPOINTS_TABLE =
  	"drop table bpecollectionpoints";

  // removed the foreign key constraint, if you are still using the old table version do
  // "db2 alter table bpecollectionpointitemtypes drop constraint fk_correlationid"
  private final String CREATE_COLLECTIONPOINTITEMTYPES_TABLE =
    "create table bpecollectionpointitemtypes "
      + "( "
      + "	seqid integer not null, "
      + "	itemtypename varchar(254), "
      + "	itemtypequantity integer "
      + ") ";
  
  private final String CREATE_COLLECTIONPOINTITEMTYPES_INDEX =
    "create index bpecps_ix "
      + "on bpecollectionpointitemtypes( seqid )";  
  
  // Drop the bpecollectionpointitemtypes table if an old version has been detected
  private final String DROP_COLLECTIONPOINTITEMTYPES_TABLE =
  	"drop table bpecollectionpointitemtypes";

  // Insert a new "request" into the CMCOLLPOINT table	
  private final String INSERT_COLLECTIONPOINTS_TABLE =
    "insert into bpecollectionpoints(seqid, correlation, itemid, timeout) values(?,?,?,?)";

  private final String INSERT_COLLECTIONPOINTITEMTYPES_TABLE =
    "insert into bpecollectionpointitemtypes(seqid, itemtypename, itemtypequantity) "
      + "values(?,?,?)";

  private final String QUERY_COLLECTIONPOINTS_TABLE =
    "select seqid "
      + "from "
      + "( "
      + " select "
      + "   p.createts, "
      + "   p.seqid, "	 
      + "   p.itemid, "
      + "   p.timeout, "
      + "   p.itemtypequantity, "
      + "   q.count, "
      + "   case "
      + "     when (q.count is null and p.itemtypequantity > 0) "
      + "          or p.itemtypequantity > q.count "
      + "       then 1 "
      + "       else 0 "
      + "   end as notready "
      + " from "
      + "   ( "
      + "     select "
      + "       x.createts, "
      + "       x.seqid, "	 
      + "       x.itemid, "
      + "       x.timeout, "
      + "       y.itemtypename, "
      + "       y.itemtypequantity "
      + "     from "
      + "       bpecollectionpoints as x, "
      + "       bpecollectionpointitemtypes as y "
      + "     where x.seqid = y.seqid "
      + "   ) as p "
      + "   left outer join "
      + "   ( "
      + "     select"
      + "       a.sourceitemid as itemid, "
      + "       c.keywordname, "
      + "       count(*) as count "
      + "     from "
      + "       icmstlinks001001 a, "
      + "       icmstitems001001 b, "
      + "       icmstnlskeywords as c "
      + "     where a.targetitemid = b.itemid "           // count items of type y.itemtypename in folder
      + "     and a.linktype = 2 "
      + "     and b.itemtypeid = c.keywordcode "
      + "     and c.keywordclass = 2 "
      + "     and c.languagecode = 'ENU' "
      + "     group by a.sourceitemid, c.keywordname "
      + "   ) as q "
      + "   on p.itemtypename = q.keywordname "
      + "   and p.itemid = q.itemid "
      + ") as r "
      + "group by createts, seqid, itemid, timeout "
      + "having sum(notready) = 0 "
      + "order by createts "
      + "with UR";
  // isolation level "uncomitted read"
  
  
  private final String QUERY_CORRELATION =
    "select "
      + "     correlation"
      + "   from "
      + "     bpecollectionpoints "
      + "  where seqid = ?";

  private final String QUERY_COLLECTIONPOINTS_EMPTY_TABLE =
    "select "
      + "     seqid, correlation, itemid"
      + "   from "
      + "     bpecollectionpoints "
      + "  where seqid not in "
      + "    (select "
      + "         seqid "
      + "       from "
      + "         bpecollectionpointitemtypes)";

  private final String QUERY_COLLECTIONPOINTS_TIMEDOUT_TABLE =
    "select "
      + "       seqid, correlation "
      + "     from "
      + "       bpecollectionpoints "
      + "     where timeout <>0 and timeout < ?";

  private final String QUERY_COLLECTIONPOINTS_OLDREQUESTS_TABLE =
    "select "
      + "       seqid, correlation "
      + "     from "
      + "       bpecollectionpoints "
      + "     where itemid not in ( select itemid from icmstitems001001 )";

  // Delete a "request" from the collection point table
  private final String DELETE_COLLECTIONPOINTS_TABLE =
    "delete from bpecollectionpoints where seqid = ?";

  // Delete all requirements for a specific "request" from the collection
  // point item types table
  private final String DELETE_COLLECTIONPOINTITEMTYPES_TABLE =
    "delete from bpecollectionpointitemtypes where seqid = ?";

}
