/*
@copyright(disclaimer)

Licensed Materials - Property of IBM
IBM DB2 Content Manager Enterprise Edition V8  (program number 5724-B19)
IBM DB2 Content Manager Express Edition V8 (program number 5724-F73)
IBM DB2 Content Manager for z/OS V8 (program number 5697-H60)
(c ) Copyright IBM Corp. 2005. All Rights Reserved.

US Government Users Restricted Rights
Use, duplication or disclosure restricted by GSA ADP Schedule
Contract with IBM Corp.

DISCLAIMER OF WARRANTIES :

Permission is granted to copy and modify this  Sample code, and to
distribute modified versions provided that both the copyright
notice,- and this permission notice and warranty disclaimer  appear in all copies and modified
versions.

THIS SAMPLE CODE IS LICENSED TO YOU AS-IS.
IBM  AND ITS SUPPLIERS AND LICENSORS  DISCLAIM
ALL WARRANTIES, EITHER EXPRESS OR IMPLIED, IN SUCH SAMPLE CODE, INCLUDING THE WARRANTY
OF NON-INFRINGEMENT AND THE IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS FOR A
PARTICULAR PURPOSE. IN NO EVENT WILL IBM OR ITS LICENSORS OR SUPPLIERS BE LIABLE FOR ANY
DAMAGES ARISING OUT OF THE USE OF  OR INABILITY TO USE THE SAMPLE CODE, DISTRIBUTION OF
THE SAMPLE CODE, OR COMBINATION OF THE SAMPLE CODE WITH ANY OTHER CODE. IN NO EVENT SHALL
IBM OR ITS LICENSORS AND SUPPLIERS BE LIABLE FOR ANY LOST REVENUE, LOST PROFITS OR DATA,
OR FOR DIRECT, INDIRECT, SPECIAL, CONSEQUENTIAL,INCIDENTAL OR PUNITIVE DAMAGES, HOWEVER
CAUSED AND REGARDLESS OF THE THEORY OF LIABILITY,-, EVEN IF IBM OR ITS LICENSORS OR SUPPLIERS
HAVE BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.

@endCopyright
*/

package com.ibm.bpe.cm.util;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;
import java.util.StringTokenizer;

import javax.ejb.CreateException;
import javax.naming.InitialContext;

import com.ibm.bpe.api.LocalBusinessFlowManager;
import com.ibm.bpe.api.LocalBusinessFlowManagerHome;
import com.ibm.bpe.api.ProcessException;
import com.ibm.bpe.api.QueryResultSet;
import com.ibm.bpe.cm.CMBPCConstants;
import com.ibm.bpe.cm.conmgmt.ManageConnection;
import com.ibm.mm.sdk.common.DKConstant;
import com.ibm.mm.sdk.common.DKDDO;
import com.ibm.mm.sdk.common.DKException;
import com.ibm.mm.sdk.common.DKFolder;
import com.ibm.mm.sdk.common.DKNVPair;
import com.ibm.mm.sdk.common.DKPidICM;
import com.ibm.mm.sdk.common.DKSequentialCollection;
import com.ibm.mm.sdk.common.DKSequentialIterator;
import com.ibm.mm.sdk.common.DKUsageError;
import com.ibm.mm.sdk.common.dkResultSetCursor;
import com.ibm.mm.sdk.server.DKDatastoreICM;
import com.ibm.websphere.bo.BOFactory;
import com.ibm.ws.extensionhelper.ExtensionHelper;
import com.ibm.ws.extensionhelper.TransactionControl;
import com.ibm.ws.extensionhelper.TxHandle;

/**
 * Implementation for a combined query of workflow (WPC) and content (CM) data.
 *  
 * The code here needs distinguish between the case when data for WPC and CM
 * is collocated or not.
 *   
 */
public class CombinedQuery
{

  /**
   * Create the object to run combined queries based on the references to related resources.
   * @param bpcDataSource A reference to the database of WPC
   * @param lsDataSource A reference to the CM library server database
   * @param bfmHome The home interface for the WPC API
   * @param bof The Business Object Factory provides the capability to create a business graph or a business object.
   * 
   */
  public CombinedQuery(LocalBusinessFlowManagerHome bfmHome, BOFactory bof)
  {    
    _bfmHome = bfmHome;
    _bof = bof;

    // Initialize TX handling
    try
    {
      InitialContext initialContext = new InitialContext();
      extHelper = (ExtensionHelper) initialContext.lookup(CMBPCConstants.EXTENSION_HELPER_REF);
      tranControl = extHelper.getTransactionControl();
    }
    catch (Exception e)
    {
      // no error handling yet but avoid anything crashes
    }
  }

  /**
   * Run a combined query for Process Choreographer and Content Manager.
   * For more details on the QueryResultSet and how to specify the clauses, also refer
   * to the Process Choreographer documentation.  
   * @param selectClauseWorkflow The requested Process Choreographer column names
   * @param flowItemTypeNameContent The item type of the item which is routed through the flow
   * This can be a document item type or a folder item type.
   * @param flowItemTypeAttributesContent A comma separated list of attributes of the flow item type
   * that should be returned in the result set. 
   * @param additionalAttributesContent If the flowItemTypeContent is a folder, this specifies an 
   * optional list of attributes of the contained documents. The list is comma separated and
   * the attributes are specified in the form <itemtype>/<attribute> 
   * @param whereClauseWorkflow The SQL where condition for Process Choreographer data
   * @param xPathContent  An XPath expression to filter the Content Manager data. The result must
   * be the flow item type. To express conditions on contained item attributes make sure
   * to use an inboundlink to get to the folder item. 	  	
   * @param orderByClauseWorkflow The order clause for Process Choreographer data   
   * @param timezoneWorkflow A timezone specification for timestamp given or returned 
   * @return The result of the combined query.
   */
  public CombinedQueryResultSet combinedQuery(
    String selectClauseWorkflow,
    String flowItemTypeNameContent,
    String flowItemTypeAttributesContent,
    String additionalAttributesContent,
    String whereClauseWorkflow,
    String xPathContent,
    String orderByClauseWorkflow,
    java.util.TimeZone timezoneWorkflow)
  {

    if (flowItemTypeNameContent == null)
    {
      throw new IllegalArgumentException("Flow item type is null");
    }

    // 
    // 1) Run WPC query
    //
    // Note that we always prepend the CM ITEMPID to the WPC query in order
    // to get the PID to do the smart join, but we don't want this to end up in
    // the attributes returned by the query, so we use an adapter that prevents
    // the ITEMPID column from being visible when we build the combined results.

    QueryResultSetAdapter resultSetBPC =
      new QueryResultSetAdapter(
        queryBPC(
          selectClauseWorkflow,
          whereClauseWorkflow,
          orderByClauseWorkflow,
          null,
          timezoneWorkflow),
        SKIP_ITEMPID_COLUMN);

    //
    // Build a list of qualifying itemIDs, associated with the QueryResultSet lines
    //
    Set itemIDs = new HashSet();
    while (resultSetBPC.next())
    {
      String itemPID = resultSetBPC.getAbsoluteString(1);
      try
      {
        DKPidICM pid = new DKPidICM(itemPID);
        itemIDs.add(pid.getItemId());
      }
      catch (DKException exc)
      {
        throw IntegrationException.wrap(exc);
      }
    }

    //
    // Tokenize the attributes for the flow item type and
    // the additional attributes. Care about null parameters - the
    // resulting list is always not null.
    //
    List flowItemTypeAttributes = new ArrayList();
    if (flowItemTypeAttributesContent != null)
    {
      StringTokenizer tok = new StringTokenizer(flowItemTypeAttributesContent, ", ");
      while (tok.hasMoreTokens())
      {
        flowItemTypeAttributes.add(tok.nextToken());
      }
    }
    List additionalAttributes = new ArrayList();
    if (additionalAttributesContent != null)
    {
      StringTokenizer tok = new StringTokenizer(additionalAttributesContent, ", ");
      while (tok.hasMoreTokens())
      {
        additionalAttributes.add(tok.nextToken());
      }
    }

    //
    // Build the list of Cm column names
    //
    List attributeNames =
      getColumnNames(flowItemTypeNameContent, flowItemTypeAttributes, additionalAttributes);

    //
    // 2) Run Cm query
    //
    HashMap resultSetCm =
      queryCm(
        flowItemTypeNameContent,
        flowItemTypeAttributes,
        additionalAttributes,
        xPathContent,
        itemIDs);

    //
    // Combine both results
    //
    CombinedQueryResultSet combinedQueryResult =
      new CombinedQueryResultSet(resultSetBPC, attributeNames, timezoneWorkflow, _bof);

    // Loop over the CM result Set and skip WPC results if required.
    if (resultSetBPC.first())
    {
      do
      {
        String itemPid = resultSetBPC.getAbsoluteString(1);
        String compareString = null;
        try
        {
          DKPidICM pid = new DKPidICM(itemPid);
          compareString = pid.getItemId() + pid.getComponentId() + pid.getVersionNumber();
        }
        catch (DKException exc)
        {
          throw new IntegrationException(exc);
        }

        List cmColumns = (List) resultSetCm.get(compareString);
        if (cmColumns != null)
        {
          combinedQueryResult.add(resultSetBPC, cmColumns);
        }
      }
      while (resultSetBPC.next());
    }

    return combinedQueryResult;
  }

  /**
   * do a query to the Process Choreographer 
   * @param selectClauseBPC
   * @param whereClauseBPC
   * @param orderByClause
   * @param threshold
   * @param timezone
   * @return the result as a ResultSet
   */
  QueryResultSet queryBPC(
    String selectClauseBPC,
    String whereClauseBPC,
    String orderByClause,
    Integer threshold,
    java.util.TimeZone timezone)
  {

    QueryResultSet resultSet = null;
    try
    {
      //    Add process attribute query part for itemPID			
      if (selectClauseBPC != null && selectClauseBPC.length() > 0)
      {
        selectClauseBPC = "PROCESS_ATTRIBUTE.VALUE AS ITEMPID, " + selectClauseBPC;
      }
      else
      {
        selectClauseBPC = "PROCESS_ATTRIBUTE.VALUE";
      }
      if (whereClauseBPC != null && whereClauseBPC.length() > 0)
      {
        whereClauseBPC =
          "("
            + whereClauseBPC
            + ") AND PROCESS_ATTRIBUTE.NAME='"
            + CMBPCConstants.PROCESS_PROPERTY_ITEMPID
            + "'";
      }
      else
        whereClauseBPC = "PROCESS_ATTRIBUTE.NAME='" + CMBPCConstants.PROCESS_PROPERTY_ITEMPID + "'";

      LocalBusinessFlowManager bfm = _bfmHome.create();
      resultSet = bfm.query(selectClauseBPC, whereClauseBPC, orderByClause, threshold, timezone);

    }
    catch (CreateException exc)
    {
      exc.printStackTrace();
    }
    catch (ProcessException exc)
    {
      exc.printStackTrace();
    }

    return resultSet;
  }

  /**
   * Run a query for the document content in Content Manager.
   * This is a follow-on step to the WPC which already returned a set of itemIDs.
   * @param flowItemTypeName The name of the item type which is routed through the flow.
   * @param flowItemTypeAttributes A comma separated list of attribute names for the flow item type.
   * @param additionalAttributes A list of attribute names in the form <itemType>/<attribute>   
   * @param xPathContent An XPath expression to filter the result set
   * @param itemIDs Set of either itemID from associated WPC processes
   * @return A Map key=itemPID, value=List of attribute values (current limitation: Strings only)
   */
  HashMap queryCm(
    String flowItemTypeName,
    List flowItemTypeAttributes,
    List additionalAttributes,
    String xPathContent,
    Set itemIds)
  {
    HashMap result = new HashMap();

    TxHandle globalTransaction = null;
    DKDatastoreICM dkDatastore = null;

    try
    {
      globalTransaction = suspendGlobalTransaction();
      dkDatastore = ManageConnection.connect();

      //
      // Concept of this combined query implementation:
      // If only attributes for the flow item type (typically, a folder) are requested,
      // we run a usual query with a projection option for the requested attributes.
      // If additional attributes are requested, we have to run an additional query
      // for each result returned to retrieve the corresponding attributes of contained
      // documents.
      // Note that we do not support the case where multiple instances of the same
      // item type are stored in a folder.
      // 
      DKNVPair[] queryOptionsFlowItemAttributes =
        buildQueryOptionForFlowItemAttributes(flowItemTypeName, flowItemTypeAttributes);
      DKNVPair[] queryOptionsAdditionalAttributes =
        buildQueryOptionForAdditionalAttributes(additionalAttributes);

      //
      // Since the number of itemId "candidates" can be huge, split it up into
      // buckets to build the ITEMID IN ( ..., ...., ...) clause.
      // 
      Object[] itemIdArray = itemIds.toArray();
      for (int i = 0; i < itemIdArray.length; i += BUCKET_SIZE)
      {
        String query = buildQuery(xPathContent, itemIdArray, i, BUCKET_SIZE);
        //
        // Run the "primary" query for the flow item types using the projection list
        // for the flow item type attributes
        //        
        System.out.println(">> Query [" + query + "]");
        dkResultSetCursor results =
          dkDatastore.execute(query, DKConstant.DK_CM_XQPE_QL_TYPE, queryOptionsFlowItemAttributes);
        DKDDO ddo = null;

        while ((ddo = results.fetchNext()) != null)
        {
          DKPidICM itemPid = (DKPidICM) ddo.getPidObject();

          // This list will store the column values in the order flowItemTypeAttributes ... additionalAttributes
          List resultColumns =
            new ArrayList(flowItemTypeAttributes.size() + additionalAttributes.size());

          for (int a = 0; a < flowItemTypeAttributes.size(); a++)
          {
            // 
            // For each flow attribute (note that it does not contain the item type name)
            // get the value from the DDO by name
            //
            String attributeValue =
              (String) ddo.getDataByName((String) flowItemTypeAttributes.get(a));
            resultColumns.add(attributeValue);
          }
          if (additionalAttributes.size() > 0)
          {
            //
            // If additional (non-flow item type) values have been requested,
            // things get expensive - run an additional query that returns
            // all items inside the folder. Use again projection.
            //

            ddo.retrieve(DKConstant.DK_CM_CONTENT_LINKS_OUTBOUND);
            // Outbound Links must be specified.

            short dataid = ddo.dataId(DKConstant.DK_CM_NAMESPACE_ATTR, DKConstant.DK_CM_DKFOLDER);

            if (dataid == 0)
              throw new IntegrationException("Flow item must be a folder if additional attributes are specified.");

            DKFolder dkFolder = (DKFolder) ddo.getData(dataid);
            DKSequentialIterator iter = (DKSequentialIterator) dkFolder.createIterator();
            HashMap attributes = new HashMap();
            dkDatastore.retrieveObjects(dkFolder); // retrieve data for DDOs in folder

            DKDDO containedDDO;
            while (iter.more())
            {
              containedDDO = (DKDDO) iter.next();
              String containedItemTypeName = containedDDO.getObjectType();
              for (dataid = 1; dataid <= containedDDO.dataCount(); dataid++)
              {
                //
                // Collect all attributes of the contained item and
                // store it in a map. Key is the itemId/attributeName string
                // which matches syntactically with the additionalAttributes list
                // content.
                //
                String containedAttributeName = containedDDO.getDataName(dataid);
                Object containedAttributeValue = containedDDO.getData(dataid);
                String itemAttributeName = containedItemTypeName + "/" + containedAttributeName;
                attributes.put(itemAttributeName, containedAttributeValue);
              }
            } // iterate over attributes
            //
            // Map the attributes in the list to the map with the retrieved values 
            //
            for (int a = 0; a < additionalAttributes.size(); a++)
            {
              String itemAttribute = (String) additionalAttributes.get(a);
              Object value = attributes.get(itemAttribute);
              resultColumns.add(value);
            }
          } // end if additional attributes are specified

          //
          // In order to correlate the WPC results with the CM results,
          // create a compare string out of the parts of an itemPID
          // Note that the itemPid.toString() cannot be used here. 
          String compareString =
            itemPid.getItemId() + itemPid.getComponentId() + itemPid.getVersionNumber();

          result.put(compareString, resultColumns);

        } // end of results loop        
        results.destroy();
      }
    }
    catch (Exception exc)
    {
      throw IntegrationException.wrap(exc);
    }
    finally
    {
      if (dkDatastore != null)
      {
        try
        {
          ManageConnection.disconnect(dkDatastore);
        }
        catch (Exception ignored)
        {
        }
      }
      if (globalTransaction != null)
      {
        resumeGlobalTransaction(globalTransaction);
      }
    }
    return result;
  }

  /**
   * Suspend an existing global transaction.
   * @return The transaction object. Use this object to resume the transaction via <code>resumeGlobalTransaction</code>.
   */
  protected TxHandle suspendGlobalTransaction()
  {
    TxHandle transaction = null;
    try
    {
      transaction = tranControl.preinvoke(true, false);
    }
    catch (javax.transaction.SystemException exc)
    {
      // Here, we can either continue processing or rethrow the exception.
      // Assuming that following JDBC calls may be successful even if the above exception was
      // thrown, we continue processing.
    }
    catch (javax.transaction.NotSupportedException exc)
    {
      // Here, we can either continue processing or rethrow the exception.
      // Assuming that following JDBC calls may be successful even if the above exception was
      // thrown, we continue processing.
    }
    return transaction;
  }

  /**
   * Resume a previously suspended global transaction.
   * @param transaction The transaction object returned by <code>suspendGlobalTransaction()</code>.
   */
  protected void resumeGlobalTransaction(final TxHandle transaction)
  {
    try
    {
      tranControl.postinvoke(transaction); // resumes previous global or local UOW
    }
    catch (javax.transaction.SystemException exc)
    {
      // Here, we can either continue processing or rethrow the exception.
      // Assuming that following JDBC calls may be successful even if the above exception was
      // thrown, we continue processing.
    }
    catch (javax.transaction.InvalidTransactionException exc)
    {
      // This exception is considered more severe than the SystemException during suspend above,
      // since we retrieved a non-null transaction object that now turns out to be invalid.
      throw IntegrationException.wrap(exc);
    }
    catch (com.ibm.ws.extensionhelper.exception.InconsistentTransactionException exc)
    {
      // This exception is considered more severe than the SystemException during suspend above,
      // since we retrieved a non-null transaction object that now turns out to be invalid.
      throw IntegrationException.wrap(exc);
    }
  }

  /**
   * Build a CM XPath query for a given xPath "filter" and a set of itemIds.
   * This function takes care about the size of the object array and does not exceed its length.
   * @param itemIdArray An array of ItemIds
   * @param startIndex The start index in the item Id array
   * @param slice The size of the slice (number of itemIds to process)
   * @return An executable query string, combined the given xPath expression and the ITEMID part.
   */
  String buildQuery(String xPathContent, Object[] itemIdArray, int startIndex, int slice)
  {
    // Create two strings  - 
    // the item id list -- ("id1","id2","id3", ...)
    // and the item id predicate @ITEMID IN("id1","id2","id3", ...)     
    StringBuffer itemIdList = new StringBuffer("(");
    for (int i = startIndex; i < startIndex + slice; i++)
    {
      if (i >= itemIdArray.length)
        break;
      if (i > 0)
        itemIdList.append(',');
      itemIdList.append('\"');
      itemIdList.append(itemIdArray[i]);
      itemIdList.append('\"');
    }
    itemIdList.append(")");
    String itemIdListPredicate = "@ITEMID IN " + itemIdList.toString();

    // Replace the substitution variables in the query expression
    String substitutedXpath = xPathContent.replaceAll(CMBPCConstants.ITEM_LIST, itemIdList.toString());
    substitutedXpath = substitutedXpath.replaceAll(CMBPCConstants.ITEM_LIST_PREDICATE, itemIdListPredicate);

    return substitutedXpath;
  }

  /**
   * Build the query options to retrieve attributes from the flow item type.
   * If no attributes are to be returned, use DK_CM_CONTENT_IDONLY to return only the IDs of the items.
   * @param flowItemTypeName The name of the item type which is routed through the flow. 
   * @param flowItemTypeAttributes A list of attribute names for the flow item type.
   * @return A DKNVPair array which can be used for a execute() call as the last query parameter.
   * @throws DKUsageError
   */
  private DKNVPair[] buildQueryOptionForFlowItemAttributes(
    String flowItemTypeName,
    List flowItemTypeAttributes)
    throws DKUsageError
  {
    DKNVPair[] options = null;
    if (flowItemTypeAttributes.size() > 0)
    {
      //
      // Attributes of the flow item type have been requested
      // Prepare a projection query option to return only these
      // attributes.
      //
      DKSequentialCollection attributeNames = new DKSequentialCollection();

      // For each attribute in the attribute specification... 
      for (int i = 0; i < flowItemTypeAttributes.size(); i++)
      {
        // Get the attribute specification of the form <item-type>/<attribute-name>
        String attributeName = (String) flowItemTypeAttributes.get(i);
        // Add this attribute name to the collection of attributes for this item type
        attributeNames.addElement(attributeName);
      }
      // Allocate an array of ordered pairs (item type, collection of attribute names for item type)
      DKNVPair[] projectionList = new DKNVPair[1];
      // Create an ordered pair to hold the item type and attribute name collection
      projectionList[0] = new DKNVPair(flowItemTypeName, attributeNames);
      // Initialize the options for the query
      options =
        new DKNVPair[] {
          new DKNVPair(DKConstant.DK_CM_PARM_PROJECTION_LIST, projectionList),
          new DKNVPair(DKConstant.DK_CM_PARM_END, null)};

    }
    else
    {
      //
      // No attributes of the flow item types have been requested
      // Prepare query options to only return the ID of the items
      //
      options =
        new DKNVPair[] {
          new DKNVPair(
            DKConstant.DK_CM_PARM_RETRIEVE,
            new Integer(DKConstant.DK_CM_CONTENT_IDONLY)),
          new DKNVPair(DKConstant.DK_CM_PARM_END, null)};
    }

    return options;
  }

  /**
   * Build the query options for any requested additional attributes (those not belonging
   * to the flow item type). 
   * @param attributeSpecifications A list of attribute names in the form <itemType>/<attribute>
   * @return A DKNVPair array which can be used for a execute() call as the last query parameter.
   * @throws DKUsageError
   */
  private DKNVPair[] buildQueryOptionForAdditionalAttributes(List attributeSpecifications)
    throws DKUsageError
  {
    HashMap itemTypes = new HashMap();
    DKNVPair[] options = null;
    // For each attribute in the attribute specification... 
    for (int i = 0; i < attributeSpecifications.size(); i++)
    {
      // Get the attribute specification of the form <item-type>/<attribute-name>
      String attributeSpecification = (String) attributeSpecifications.get(i);
      // Get the location of the delimiter between the item type and attribute name
      int delimiter = attributeSpecification.indexOf('/');
      if (delimiter == -1)
        throw new IntegrationException("Missing delimiter. Attribute name must be of form \"item-type/attribute-name\"");
      // Get the text up to the delimiter as the item type name
      String itemTypeName = attributeSpecification.substring(0, delimiter);
      // Get the text after the delimiter as the attribute name (which may contain more slashes)
      String attributeName = attributeSpecification.substring(delimiter + 1);
      // Get the attribute names (for this item type) that we've seen so far
      DKSequentialCollection attributeNames = (DKSequentialCollection) itemTypes.get(itemTypeName);
      // If this is the first attribute name for this item type
      if (attributeNames == null)
      {
        // Allocate a new collection to hold the attribute names for this item type
        attributeNames = new DKSequentialCollection();
        // Save the collection so it can be retrieved later by item type name
        itemTypes.put(itemTypeName, attributeNames);
      }
      // Add this attribute name to the collection of attributes for this item type
      attributeNames.addElement(attributeName);
    }
    // Allocate an array of ordered pairs (item type, collection of attribute names for item type)
    DKNVPair[] projectionList = new DKNVPair[itemTypes.size()];
    int projectionListIndex = 0;
    // For each unique item type in the attribute specification (note that we have now coalesced
    // multiple possible attribute specifications into one per item type)
    for (Iterator itemTypeIterator = itemTypes.keySet().iterator(); itemTypeIterator.hasNext();)
    {
      // Get the name of the item type
      String itemTypeName = (String) itemTypeIterator.next();

      // Get the collection of attribute names for this item type
      DKSequentialCollection attributeNames = (DKSequentialCollection) itemTypes.get(itemTypeName);

      // Create an ordered pair to hold the item type and attribute name collection
      projectionList[projectionListIndex++] = new DKNVPair(itemTypeName, attributeNames);
    }
    // Initialize the options for the query
    options =
      new DKNVPair[] {
        new DKNVPair(DKConstant.DK_CM_PARM_PROJECTION_LIST, projectionList),
        new DKNVPair(DKConstant.DK_CM_PARM_END, null)};
    return options;
  }

  private List executeQueryCm(
    String query,
    String flowItemType,
    List flowItemTypeAttributes,
    List additionalAttributes)
  {
    List resultColumns = new ArrayList();
    //
    // If flow item types are requested, run projection to get only the
    // flow item type attributes returned.
    // If no flow item types are requested, only return the IDs
    //

    return resultColumns;

  }

  /**
   * Build the list of (display) column names for the requested attributes on the flow item type
   * and the additional attributes.
   * Note that the additional attributes are already of theform <itemType>/<attribute>
   * @param flowItemTypeName The name of the item type which is routed through the flow
   * @param flowItemTypeAttributes A list of attributes of the flow item type
   * @param additionalAttributes A list of additional attributes of the form <itemType>/<attribute>
   * @return A list of strings, each representing the name of an attribute  
   */
  List getColumnNames(
    String flowItemTypeName,
    List flowItemTypeAttributes,
    List additionalAttributes)
  {
    List columnNames = new ArrayList();
    for (int i = 0; i < flowItemTypeAttributes.size(); i++)
    {
      columnNames.add(flowItemTypeName + "/" + flowItemTypeAttributes.get(i));
    }
    for (int i = 0; i < additionalAttributes.size(); i++)
    {
      columnNames.add(additionalAttributes.get(i).toString());
    }
    return columnNames;
  }

  // Some variables for allowing local transactions
  ExtensionHelper extHelper = null;
  TransactionControl tranControl = null;

  /**
   * Reference to the WPC generic API
   */
  private LocalBusinessFlowManagerHome _bfmHome = null;
  
  /**
   * Reference to the BO factory
   */
  private BOFactory _bof = null;

  private final int BUCKET_SIZE = 50;
  // Note that the number of '?'s in this String must match with BUCKET_SIZE
  private final String INCLAUSE =
    "IN(?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?)";
  private final int INCLAUSE_START_PARM = 3;
  private final int INCLAUSE_LENGTH_PARM = 2;

  private final static int SKIP_ITEMPID_COLUMN = 1;

}
