/*
@copyright(disclaimer)

Licensed Materials - Property of IBM
IBM DB2 Content Manager Enterprise Edition V8  (program number 5724-B19)
IBM DB2 Content Manager Express Edition V8 (program number 5724-F73)
IBM DB2 Content Manager for z/OS V8 (program number 5697-H60)
(c ) Copyright IBM Corp. 2005. All Rights Reserved.

US Government Users Restricted Rights
Use, duplication or disclosure restricted by GSA ADP Schedule
Contract with IBM Corp.

DISCLAIMER OF WARRANTIES :

Permission is granted to copy and modify this  Sample code, and to
distribute modified versions provided that both the copyright
notice,- and this permission notice and warranty disclaimer  appear in all copies and modified
versions.

THIS SAMPLE CODE IS LICENSED TO YOU AS-IS.
IBM  AND ITS SUPPLIERS AND LICENSORS  DISCLAIM
ALL WARRANTIES, EITHER EXPRESS OR IMPLIED, IN SUCH SAMPLE CODE, INCLUDING THE WARRANTY
OF NON-INFRINGEMENT AND THE IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS FOR A
PARTICULAR PURPOSE. IN NO EVENT WILL IBM OR ITS LICENSORS OR SUPPLIERS BE LIABLE FOR ANY
DAMAGES ARISING OUT OF THE USE OF  OR INABILITY TO USE THE SAMPLE CODE, DISTRIBUTION OF
THE SAMPLE CODE, OR COMBINATION OF THE SAMPLE CODE WITH ANY OTHER CODE. IN NO EVENT SHALL
IBM OR ITS LICENSORS AND SUPPLIERS BE LIABLE FOR ANY LOST REVENUE, LOST PROFITS OR DATA,
OR FOR DIRECT, INDIRECT, SPECIAL, CONSEQUENTIAL,INCIDENTAL OR PUNITIVE DAMAGES, HOWEVER
CAUSED AND REGARDLESS OF THE THEORY OF LIABILITY,-, EVEN IF IBM OR ITS LICENSORS OR SUPPLIERS
HAVE BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.

@endCopyright
*/

package com.ibm.bpe.cm.util;

import java.util.ArrayList;
import java.util.List;
import java.util.TimeZone;

import com.ibm.bpe.api.QueryColumnInfo;
import com.ibm.bpe.cm.CMBPCConstants;
import com.ibm.websphere.bo.BOFactory;
import commonj.sdo.DataObject;

/**
 * Combine a WPC result set with a CM query result set.
 * The constructor is used to gather sizes and meta information,
 * the <code>add</code> function is used to add new rows to the result set.
 * An object of this class is returned as a result of a combined query.<br>
 * Note: this version assumes that the CM contribution are only Strings.
 * Numerical values are translated into Strings.
 */
public class CombinedQueryResultSet
{
	
	
  /**
   * Create a new combined result set for queries.
   * @param resultSetBPC A result set from a WPC query
   * @param columnNamesCM A list of Strings with the columns names (headings) for the CM part
   * @param timezone (optional) The timezone if the query expression had TS() attributes.
   */
  public CombinedQueryResultSet(
    QueryResultSetAdapter resultSetBPC,
    List columnNamesCM,
    TimeZone timezone,
	BOFactory bof)
  {
    // The number of columns from the BPC, CM part and the total number of columns		
    _numberColumnsBPC = resultSetBPC.numberColumns();
    _numberColumnsCM = columnNamesCM.size();
    _numberColumns = _numberColumnsBPC + _numberColumnsCM;

    _timezone = timezone;
    
    _bof = bof;

    if (_bof == null) {  // Create a list object for use in a non-SCA environment
        _queryResultSetList = new ArrayList();
    }
    else {     // Use the BO factory to create the Data object for the query result list. This object stores an array of result rows
        _queryResultSetBO = _bof.create(CMBPCConstants.CMSERVICES_NAMESPACE,CMBPCConstants.BO_QUERYRESULTSET);       	
    }

    // Assign the variables for the display names (headers) and 
    // information about the columns (type, name, etc.)
    _columnDisplayNames = new String[_numberColumns];
    _selectedColumns = new ArrayList();
    // Copy information for the WPC piece
    for (int i = 0; i < _numberColumnsBPC; i++)
    {
      String columnDisplayName = resultSetBPC.getColumnDisplayName(i + 1);
      short columnType = resultSetBPC.getColumnType(i + 1);
      _columnDisplayNames[i] = columnDisplayName;
      QueryColumnInfo columnInfo = new QueryColumnInfo(columnDisplayName, columnType, true);
      _selectedColumns.add(columnInfo);
    }
    // Copy information for the CM piece
    for (int i = _numberColumnsBPC; i < _numberColumns; i++)
    {
      String columnDisplayName = (String) columnNamesCM.get(i - _numberColumnsBPC);
      short columnType = QueryColumnInfo.TYPE_STRING;
      _columnDisplayNames[i] = columnDisplayName;
      QueryColumnInfo columnInfo = new QueryColumnInfo(columnDisplayName, columnType, true);
      _selectedColumns.add(columnInfo);
    }
  }

  /**
   * Add a WPC resultSet row and a list of CM columns to the combined result set.
   * @param columnsBPC A row from a WPC result set (more precisely - the entire result set with the current 
   * cursor pointing to a particular row)
   * @param columnsCM A list of Strings, representing the associated CM column values.
   */
  public void add(QueryResultSetAdapter columnsBPC, List columnsCM)
  {    
  	List rows, columns;
  	DataObject queryResultSetRow = null;
  	// 
    // Allocate a new result entry for the resulting columns
  	//
  	if (_bof == null) {
  		columns = new ArrayList();
    }
  	else {
  		queryResultSetRow = _bof.create(CMBPCConstants.CMSERVICES_NAMESPACE, CMBPCConstants.BO_QUERY_RESULT_SET_ROW);
  	  	columns = queryResultSetRow.getList(CMBPCConstants.BOATT_QUERY_RESULT_SET_ROW_COLUMNS);
  	}
  	
    //
    // Assign the WPC columns to resultEntry[] - note that IDs are handled differently
    // since QueryResultSetImpl in the WPC code has to care about Informix which cannot store
    // binaries. Thus, get the OID object and store it in the entry array.
    //
    for (int i = 0; i < _numberColumnsBPC; i++)
    {
        QueryColumnInfo ci = (QueryColumnInfo) _selectedColumns.get(i);
        if (ci.getType() == QueryColumnInfo.TYPE_ID) {
    		columns.add( columnsBPC.getOID(i + 1).toString());
    	}
    	else {
        	Object obj = columnsBPC.getObject(i + 1);
          	columns.add( (obj == null) ? null : obj.toString() );            		
    	}
    }
    //
    // Assign the CM columns to resultEntry[] - start the index at _numberColumnsBPC
    // and keep in mind - everything in the List should be Strings.
    // Just to be sure, call a String.valueOf() but care about null values.
    //
    for (int i = _numberColumnsBPC; i < _numberColumns; i++)
    {
      Object columnValue = columnsCM.get(i - _numberColumnsBPC);
      if (columnValue == null)
      {
      	columns.add( null );    
      }
      else
      {
      	columns.add( String.valueOf(columnValue));        
      }
    }
    
    if (_bof == null) {
        _queryResultSetList.add( columns ); // columns is a List of columns
    }
    else {
    	rows = _queryResultSetBO.getList(CMBPCConstants.BOATT_QUERY_RESULT_SET_ROW_ROWS);
    	rows.add( queryResultSetRow );	// queryResultSetRow is the Data Object representing a list of rows
    }
    
  }


  /**
    * The number of the columns in the result.
    * The <tt>columnIndex</tt> parameter in the following calls
    * must be in the range <tt>1 .. numberColumns()</tt>.
    *
    * @return Number of columns in the result.
    */
  public final int numberColumns()
  {
    return _selectedColumns.size();
  }

  /**
   * Name of the column (header).
   * Use this name to display the column title.
   * The value of the column header is independent of the current result entry
   * position.
   *
   * @param columnIndex the first column is 1, the second is 2, ...
   * @return the header name of the column
   * @see #getTableDisplayName    
   */
  public final String getColumnDisplayName(int columnIndex)
  {
    if (columnIndex < 1 || columnIndex > numberColumns())
    {
      throw new IndexOutOfBoundsException(Integer.toString(columnIndex));
    }
    if (_columnDisplayNames == null)
    {
      // no results were added to the result list, therefore
      // no label names from the meta info was added
      // return the column name
      return ((QueryColumnInfo) _selectedColumns.get(columnIndex - 1)).getColumnName();

    }
    //
    // The reason why the column display name is usually retrieved from the result
    // set rather than from the QueryColumnInfo are column alias names that 
    // cannot be retrieved from the object but only from the result set meta data
    // This, however, works only if there is at least one row in the result set
    //
    return _columnDisplayNames[columnIndex - 1];
  }

  /**
   * Return a Data Object representation of the query result set.
   * The SDO has a list of rows (QueryResultSetRows) which contains
   * again a list of columns.
   * @return The data object with the query result set.
   */
  public DataObject getDataObject()
  {
  	return _queryResultSetBO;
  }

  /**
   * Return a List representation of the query result set.
   * The list comprises rows that are lists of columns.
   * @return The data object with the query result set.
   */
  public List getList()
  {
  	return _queryResultSetList;
  }


  // A list of QueryColumnInfo objects describing
  // the selected columns
  private List _selectedColumns = null;
  // An array of the display names for the columns
  private String[] _columnDisplayNames = null;

  // List of resultSet entries, one entry represents one row,
  // each entry is an array of Objects
  private DataObject _queryResultSetBO = null;
  private List _queryResultSetList = null;

  //
  // The timezone that was specified in the query() call and
  // that is now used to return timestamp values
  // If this is null, UTC time was requested.
  //
  private TimeZone _timezone = null;

  private final TimeZone _utcTimeZone = TimeZone.getTimeZone("UTC");

  private int _numberColumns = 0;
  private int _numberColumnsBPC = 0;
  private int _numberColumnsCM = 0;
  
  private BOFactory _bof = null;

  static final long serialVersionUID = 1L;
}
