/*
@copyright(disclaimer)

Licensed Materials - Property of IBM
IBM DB2 Content Manager Enterprise Edition V8  (program number 5724-B19)
IBM DB2 Content Manager Express Edition V8 (program number 5724-F73)
IBM DB2 Content Manager for z/OS V8 (program number 5697-H60)
(c ) Copyright IBM Corp. 2005. All Rights Reserved.

US Government Users Restricted Rights
Use, duplication or disclosure restricted by GSA ADP Schedule
Contract with IBM Corp.

DISCLAIMER OF WARRANTIES :

Permission is granted to copy and modify this  Sample code, and to
distribute modified versions provided that both the copyright
notice,- and this permission notice and warranty disclaimer  appear in all copies and modified
versions.

THIS SAMPLE CODE IS LICENSED TO YOU AS-IS.
IBM  AND ITS SUPPLIERS AND LICENSORS  DISCLAIM
ALL WARRANTIES, EITHER EXPRESS OR IMPLIED, IN SUCH SAMPLE CODE, INCLUDING THE WARRANTY
OF NON-INFRINGEMENT AND THE IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS FOR A
PARTICULAR PURPOSE. IN NO EVENT WILL IBM OR ITS LICENSORS OR SUPPLIERS BE LIABLE FOR ANY
DAMAGES ARISING OUT OF THE USE OF  OR INABILITY TO USE THE SAMPLE CODE, DISTRIBUTION OF
THE SAMPLE CODE, OR COMBINATION OF THE SAMPLE CODE WITH ANY OTHER CODE. IN NO EVENT SHALL
IBM OR ITS LICENSORS AND SUPPLIERS BE LIABLE FOR ANY LOST REVENUE, LOST PROFITS OR DATA,
OR FOR DIRECT, INDIRECT, SPECIAL, CONSEQUENTIAL,INCIDENTAL OR PUNITIVE DAMAGES, HOWEVER
CAUSED AND REGARDLESS OF THE THEORY OF LIABILITY,-, EVEN IF IBM OR ITS LICENSORS OR SUPPLIERS
HAVE BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.

@endCopyright
*/

package com.ibm.bpe.cm.conmgmt;

import java.security.AccessController;
import java.security.PrivilegedExceptionAction;

import com.ibm.mm.sdk.server.DKDatastoreICM;

/**
 * Functions to access the Content Manager API.
 * This includes static functions to connect / disconnect to CM, read the
 * settings from bpecm.properties and get document content for a given itemPid.
 */
public class ManageConnection
{
  public ManageConnection()
  {
  }
	
  /**
   * Connect to Content Manager using the userid and password specified in the properties file.
   * Typically, this will connect using the icmadmin userid (or any other user with
   * admin authority) but not the userid icmconct.
   * @return A content manager connection
   * @throws ConnectionException If the given parameters did not allow to connect.
   */
	public static DKDatastoreICM connect() throws ConnectionException
	{
		return connect(null);
	}

 /**
   * Connect to Content Manager using the userid and password specified in the properties file.
   * Typically, this will connect using the icmadmin userid (or any other user with
   * admin authority) but not the userid icmconct.
   * @param Optional connect string. May be null if no connect string is needed
   * @return A content manager connection
   * @throws ConnectionException If the given parameters did not allow to connect.
   */
  public static DKDatastoreICM connect(final String connectstring) throws ConnectionException
  {
    DKDatastoreICM dkDatastore = null;

    try
    {
      dkDatastore = (DKDatastoreICM) AccessController.doPrivileged(new PrivilegedExceptionAction()
      {
        public Object run() throws ConnectionException
        {
          DKDatastoreICM dkDatastore = null;
          //
          // Lookup the userid and password from the properties file (password may be encrypted) 
          //
          String uid = Environment.getProperty(Environment.BPE_CM_CONNECT_UID);
          if (uid == null)
          {
            throw new ConnectionException(
              "Cannot find userid for connecting to CM DB (property name is "
                + Environment.BPE_CM_CONNECT_UID
                + ")");
          }
          String pwd =
            Environment.getPasswd(Environment.BPE_CM_CONNECT_PWD, Environment.BPE_CM_PWD_ENCRYPT);
          if (pwd == null)
          {
            throw new ConnectionException(
              "Cannot find password for connecting to CM DB (property name is "
                + Environment.BPE_CM_CONNECT_PWD
                + ")");
          }

          try
          {
            dkDatastore = connect(uid, pwd, connectstring);
          }
          catch (Exception e)
          {
            // If alias has been null
            throw new ConnectionException(e);
          }
          return dkDatastore;
        }
      });
    }
    catch (Exception e)
    {
        throw ConnectionException.wrap(e);
    }
    return dkDatastore;
  }

  /**
   * Connect to Content Manager with the given userid.
   * This function is as well used by the WAS User Registry implementation to check userid/password
   * during logon.
   * @param userId The userid for a Content Manager connection
   * @param password The password for the userId
   * @return A content manager connection
   * @throws ConnectionException If the given parameters did not allow to connect.
   */
  public static DKDatastoreICM connect(final String userId, final String password, final String connectstring)
    throws ConnectionException
  {
    DKDatastoreICM dkDatastore = null;

    try
    {
      dkDatastore = (DKDatastoreICM) AccessController.doPrivileged(new PrivilegedExceptionAction()
      {
        public Object run() throws ConnectionException
        {
          DKDatastoreICM dkDatastore = null;

          try
          {
            // 
            // Connect, using the given userId, password, database name and schema name
            // TRUSTEDUSERLOGON needs to be disabled to ensure password checking. This is especially important
          	// when using the custom ContentManagerWebSphereUserRegitry.
            //
            dkDatastore = new DKDatastoreICM();
            StringBuffer sb = new StringBuffer();
            sb.append("TRUSTEDUSERLOGON=FALSE;SCHEMA=");
            sb.append(getSchemaName());
            if (connectstring != null) {
            	sb.append(';');
            	sb.append(connectstring);
            }
            dkDatastore.connect(getDatastoreName(), userId, password, sb.toString());
          }
          catch (Exception exc)
          {
            throw new ConnectionException(exc);
          }
          return dkDatastore;
        }
      });
    }
    catch (Exception e)
    {
      throw ConnectionException.wrap(e);
    }
    return dkDatastore;
  }

  /**
   * Disconnect from a data store
   * @param dkDatastore A connection object retrieved from connect()
   * @throws ConnectionException In case the disconnect failed
   */
  public static void disconnect(final DKDatastoreICM dkDatastore) throws ConnectionException
  {
    try
    {
      AccessController.doPrivileged(new PrivilegedExceptionAction()
      {
        public Object run() throws ConnectionException
        {
          try
          {
            if (dkDatastore != null)
            {
              dkDatastore.disconnect();
              dkDatastore.destroy();
            }
          }
          catch (Exception exc)
          {
            throw new ConnectionException(exc);
          }
          return null;
        }
      });
    }
    catch (Exception e)
    {
      throw ConnectionException.wrap(e);
    }
  }


  /**    
   * @return The Content manager data store name (e.g. ICMNLSDB)
   */
  public static String getDatastoreName()
  {
    if (_datastoreName == null)
    {
      synchronized (DATASTORE_NAME_PROPERTY)
      {
        _datastoreName = Environment.getProperty(DATASTORE_NAME_PROPERTY);
        if (_datastoreName == null)
        {
          throw new ConnectionException(
            "Set "
              + DATASTORE_NAME_PROPERTY
              + " in bpecm.properties to the datastore name (e.g. ICMNLSDB).");
        }
      }
    }
    return _datastoreName;
  }

  /**   
   * @return The schema name for the Content Manager connect string SCHEMA=xxxx
   */
  public static String getSchemaName()
  {
    if (_schemaName == null)
    {
      synchronized (SCHEMA_NAME_PROPERTY)
      {
        _schemaName = Environment.getProperty(SCHEMA_NAME_PROPERTY);
        if (_schemaName == null)
        {
          throw new ConnectionException(
            "Set "
              + SCHEMA_NAME_PROPERTY
              + " in bpecm.properties to the schema name for a Content Manager connection (e.g. ICMADMIN)");
        }
      }
    }
    return _schemaName;
  }

  
  // Name of the data store and its corresponding property in bpecm.properties
  private static String _datastoreName = null;
  private static final String DATASTORE_NAME_PROPERTY = "ContentManagerDatastoreName";

  // Name of the schema name its corresponding property in bpecm.properties
  private static String _schemaName = null;
  private static final String SCHEMA_NAME_PROPERTY = "ContentManagerSchema";

  // Index numbers for columns in returned attribute tables. Must be kept in order of initialization.
  public final static int ITEM_TYPE_COLUMN = 0;
  public final static int PID_STRING_COLUMN = 1;
  public final static int ATTRIBUTE_NAME_COLUMN = 2;
  public final static int ATTRIBUTE_VALUE_COLUMN = 3;

}
